package ElchemeaProgram;
#$Id: ElchemeaProgram.pm,v 1.44 2012/05/30 07:51:53 sofc Exp $
####################################################################################
##
##
##      Package ElchemeaProgram
##
##
##  This module is intended for programmed sequences of impedance, potential sweeps
##  or chrono potentiometry/amperometry. It is used by the grahical interface to Elchemea
##  and can be used independendtly as well.
##
##
##  The constructor can be called in two ways:
##
##  $id = ElchemeaProgram->new($username);
##
##  $id = ElchemeaProgram->new($username,'test');
##
##  If the test argument is specified (any second argument will work similarily)
##  all the standard member function returns immediately and the destructor outputs the 
##  expected total runtime for the whole program if it were to be called without the
##  test argument. Thus if the module is used independently of the standard user
##  interface (or programs are manually edited), the programmer must make sure that all
##  non-member function calls are protected by a test guard like the one shown below:
##
##  ----\begin{verbatim}
##  ----Member function work here
##  ----if (!$id->test)
##  ----{
##  ----    Do non member function work here
##  ----}
##  ----More member function work here
##  ----\end{verbatim}
##
##  This is especially improtant if tha manual edit option is used in the graphical
##  Elchemea interface, as the interface calls the program with the test argument to
##  obtain the expected runtime in order to display this to the user!
##  All member functions include a test guard where it is nescesarry.
##  
##  A ElchemeaProgram instance includes two counters, one which is incremented for each
##  member function entry, and one for each member function success. The destructor 
##  appends a line the the users programlog file indication how many commands were 
##  attempted and how may were successfull. Note that in case the program is prematurely
##  ended (by removing the program lockfile) the two numbers will differ!
##
##  The member functions of an ElchemeaProgram instance are:
##
##  $id->test()          Returns true if the test argument was supplied to the constructor
##  $id->fra()           Returns the frequency response analyser object (refer FDEV.pm for
##                       details).
##  $id->pot()           Returns the potentiostat object (refer FDEV.pm for details).
##  $id->debug()         Sets or gets the debug level
##  $id->user()          Returns the user object (created from the username argument to 
##                       the constructor, refer ElchemeaUser.pm for details
##  $id->set_lock()      Sets the lockfile preventing other ElchemeaPrograms from running. 
##                       Automatically called by the constructor if not in test mode.
##  $id->is_locked()     Returns true if an non-test ElchemeaProgram is running.
##                       This function is used to check if a program should terminate 
##                       prematurely, if so, the member functions returns imediately
##                       (as had the test parameter been specified) and the success counter
##                       is not incremented.
##  $id->unlock()        Removes the lockfile, automatically called by the destructor for
##                       non-test programs.
##  $id->end_all()       Stops all measurements started by the program.
##  $id->end_chrono()    Stops any running chrono amperometry/potentiometry, including
##                       forked programs
##  $id->CGI_client(@)   TCP-IP client for calling the CGI-server.
##  $id->socket_client($$@)
##                       Generic TCP-IP client. usage: socket_client($host,$port,@args)
##  $id->mail($$)        Email client for sending mail, usage: mail($email,$message,[opt $subj])
##  $id->init_fra()      Initialises the frequency response analyser
##  $id->init_potentiostat()
##                       Initialises the potentiostat
##  $id->reset_potentiostat
##                       Resets the potentiostat.
##  $id->chrono($)       Runs a chrono amperometry/potentiometry. usage:
##                       chrono($session,[opt $time],[opt $polarisation]). If the polarisation 
##                       is not specified, the polarisation in the user configuration is used
##                       and if the time is 0 the process forks and the started chrono runs
##                       untill program exit (allows for background chrono measurements).
##                       If no time or polarisation is specified, the settings from the session
##                       configuration are used.
##  $id->potsweep($)     Runs a potential sweep based on specified session configuration.
##  $id->multi_imp($$$   Runs a series of impedance measuremtns based on specified session
##                       configuration. Usage: multi_imp($session,$number,$minutes_inbetween).
##  $id->impedance()     Runs a single impedance measurement based on specified session 
##                       configuration.
##  $id->proglog($)      Writes the specified string to the users programlog.
##  $id->change_channel($)
##                       Changes the multiplexer channel to the specified channel number
##                       (Only valid if a multiplexer is connected to the Elchemea system)
##  $id->get_multiplex_channel()
##                       Returns the channel selected by the multiplexer (see above)
##  $id->wait($)         Waits for the specified number of minutes.
##  $id->temp_and_ramp($$)
##                       Sets the temperature controler setpoint and ramprate, only valid
##                       If a temprature controler is connected to the Elchemea system.
##
##
############################################################################################

use strict;
use warnings;
use ElchemeaHeader;
use Elchemea;
use ElchemeaUser;
use SemaforeFile;
use Eurotherm;
use Sys::Hostname;
use Fcntl qw(:DEFAULT :flock);
use Debug;
use Time::HiRes qw(sleep time);
our @ISA = qw(Debug);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
    unless (-e $multiplex_channel_file)
    {
        open MYOUT,'>',$multiplex_channel_file or CORE::die "Could not open '$multiplex_channel_file' for writing $!";
        close MYOUT;
        chmod 0666, $multiplex_channel_file;
    }
}

return 1;

sub new
{
    my $class = shift;
    my $self = $class->SUPER::new();
    $self->{'_count'} = 0;
    $self->{'_success'} = 0;
    $self->{'_start_ok'} = 0;
    $self->{'_multiplex'} = SemaforeFile->new($multiplex_channel_file);
    $self->{'_user'} = ElchemeaUser->new(shift);
    $self->{'_pot'} = Elchemea::new_PSTAT(Elchemea::get_config_value('equipment','potentiostat'),Elchemea::get_config_value('equipment','potentiostat_address'));
    $self->{'_fra'} = Elchemea::new_FRA(Elchemea::get_config_value('equipment','FRA'),Elchemea::get_config_value('equipment','FRA_address')); 
    $self->{'_chronoforever'} = 0;
    $self->{'_start_ok'} = 1;
    $self->{'_test'} = 0;
    $self->{'_test'} = 1 if (@_);
    $self->{'_totaltime'} = 0;
    bless $self,$class;
    unless ($self->{'_test'})
    {
	if ($self->is_locked)
	{
	    $self->user->proglog("Program not started as semafore is already set!");
	    exit;
	}
	$self->set_lock;
	$self->user->proglog('Program started');
    }
    return $self;
}

sub test()
{
    my $self = shift;
    return $self->{'_test'};
}

sub fra()
{
    my $self = shift;
    return $self->{'_fra'};
}

sub pot()
{
    my $self = shift;
    return $self->{'_pot'};
}

sub debug
{
    my $self = shift;
    $self->SUPER::debug(@_);
    if (@_)
    {
        $self->{'_multiplex'}->debug($self->{'_debug'});
        $self->{'_user'}->debug($self->{'_debug'});
    }
    return $self->{'_debug'};
}

sub user()
{
    my $self = shift;
    return $self->{'_user'};
}

sub set_lock()
{
    my $self = shift;
    return if ($self->{'_test'});
    $self->writedebug("Setting lockfile $programlock");
    return if (-e $programlock);
    open MYOUT,'>',$programlock or $self->die("Could not open file $programlock for writing $!");
    close MYOUT;
    chmod 0666,$programlock;
}

sub is_locked()
{
    my $self = shift;
    return if ($self->{'_test'});
    return -e $programlock;
}

sub unlock()
{
    my $self = shift;
    return if ($self->{'_test'});
    $self->writedebug("Removing lockfile $programlock");
    unlink $programlock if (-e $programlock);
}

sub end_all()
{
    my $self = shift;
    return if ($self->{'_test'});
    $self->{'_count'}++;
    return unless ($self->is_locked);
    $self->writedebug("Stopping all programs");
    $self->user->proglog("Stopping all programs!");
    &Elchemea::unlock_measure();
    $self->end_chrono;
    $self->{'_success'}++;
}

sub end_chrono()
{
    my $self = shift;
    return if ($self->{'_test'});
    $self->{'_count'}++;
    return unless ($self->is_locked);
    $self->{'_chronoforever'} = 0;
    $self->writedebug("Stopping chrono");
    my $string = "Stopping any chrono amperometry/potentiometry";
    $self->user->proglog($string);
    my $file = $chrono_runforever_semafore;
    unlink $file if (-e $file);
    $self->{'_success'}++;
}

sub CGI_client(@)
{
    my $self = shift;
    return if ($self->{'_test'});
    $self->{'_count'}++;
    return unless ($self->is_locked);
    my $res = undef;
    eval {
	$res = Elchemea::CGI_client(@_);
    };
    if ($@)
    {
	$self->user->proglog($@);
	return undef;
    }
    $self->{'_success'}++;
    return $res;
}

sub socket_client($$@)
{
    my $self = shift;
    my $host = shift;
    my $port = shift;
    return if ($self->{'_test'});
    $self->{'_count'}++;
    return unless ($self->is_locked);
    $self->die("Host or port not specified for socket_client") unless ($host && $port);
    my $res = undef;
    eval {
	$res = Elchemea::socket_client($host,$port,@_);
    };
    if ($@)
    {
	$self->user->proglog($@);
	return undef;
    }
    $self->{'_success'}++;
    return $res;
}

sub mail()
{
    my $self = shift;
    return if ($self->{'_test'});
    my $email = shift;
    my $message = shift;
    $self->{'_count'}++;
    return unless ($self->is_locked);
    return unless ($email =~ /\w+\@\w+/ && $message);
    my $hostname = hostname;
    my $subject = 'Automatic message from Elchemea user'.$self->{'_user'}->name().' on '.$hostname;
    $subject = shift if (@_);
    my $res = undef;
    eval {
	$res = Elchemea::mail($email,$message,$subject);
    };
    if ($@)
    {
	$self->user->proglog($@);
	return undef;
    }
    $self->{'_success'}++;
    return $res;
}

sub init_fra()
{
    my $self = shift;
    return if ($self->{'_test'});
    $self->{'_count'}++;
    return unless ($self->is_locked);
    my $fra = Elchemea::new_FRA(Elchemea::get_config_value('equipment','FRA'),Elchemea::get_config_value('equipment','FRA_address'));
    unless ($fra->virtual)
    {
	my $locked = Elchemea::is_locked_device;
	$self->writedebug("Warning: Device appears to be used by an other process, initialising could result in undefined behaviour!") if ($locked);
	Elchemea::lock_device() unless ($locked);
	$self->user->proglog("Initialising FRA: ".$fra->type);
	sleep(0.3); ##Allow other programs to fish and clean up ###
	$fra->init();
	sleep(0.3);
	Elchemea::unlock_device() unless ($locked);
    }
    $self->{'_success'}++;
}

sub init_potentiostat()
{
    my $self = shift;
    return if ($self->{'_test'});
    $self->{'_count'}++;
    return unless ($self->is_locked);
    my $pot = Elchemea::new_PSTAT(Elchemea::get_config_value('equipment','potentiostat'),Elchemea::get_config_value('equipment','potentiostat_address'));
    unless ($pot->virtual)
    {
	my $locked = Elchemea::is_locked_device;
	$self->writedebug("Warning: Device appears to be used by an other process, initialising could result in undefined behaviour!") if ($locked);
	Elchemea::lock_device() unless ($locked);
	$self->user->proglog("Initialising potentiostat: ".$pot->type);
	sleep(0.3); ##Allow other programs to fish and clean up ###
	$pot->init();
	sleep(0.3);
	Elchemea::unlock_device() unless ($locked);
    }
    $self->{'_success'}++;
}

sub reset_potentiostat()
{
    my $self = shift;
    return if ($self->{'_test'});
    $self->{'_count'}++;
    return unless ($self->is_locked);
    my $pot = Elchemea::new_PSTAT(Elchemea::get_config_value('equipment','potentiostat'),Elchemea::get_config_value('equipment','potentiostat_address'));
    unless ($pot->virtual)
    {
	my $locked = Elchemea::is_locked_device;
	$self->writedebug("Warning: Device appears to be used by an other process, resetting could result in undefined behaviour!") if ($locked);
	Elchemea::lock_device() unless ($locked);
	$self->user->proglog("Resetting potentiostat ".$pot->type." to OCV conditions");
	sleep(0.3); ##Allow other programs to fish and clean up ###
	$pot->reset();
	sleep(0.3);
	Elchemea::unlock_device() unless ($locked);
    }
    $self->{'_success'}++;
}

sub chrono($)
{
    my $self = shift;
    if ($self->{'_test'})
    {
	my $session = shift;
	my $time = undef;
	if (@_)
	{
	    $time = shift;
	    $time = $time * 60;  ## time in seconds ###
	}
	$time = 1 unless ($time);
	$self->{'_totaltime'} += $time * 60 if ($time);
	return;
    }
    $self->{'_count'}++;
    return unless ($self->is_locked);
    $self->writedebug("Starting chrono");
    my $session = shift;
    $session = $self->user->session unless (defined($session));
    my $time = undef;
    if (@_)
    {
	$time = shift;
	$time = $time * 60;  ## time in seconds ###
    }
    if (defined($time) && $time == 0)
    {
	### Only allow one baseline chrono at a time
	if ($self->{'_chronoforever'})
	{
	    $self->user->proglog("Baseline chrono potentiometry/amperometry not started as one baseline measurement is already running!");
	    return 0;
	}
	$self->{'_chronoforever'} = 1;
    }
    my $string = "Starting chrono potentiometry/amperometry";
    my $pol = undef;
    if (@_)
    {
	### Setting polarisation explicitly and not from configuration ###
	$pol = shift;
	$string .= ' using polarisation '.$pol;
    }
    $self->user->proglog($string);
    my $command = join(' ','env DISPLAY=dumb','/usr/local/bin/elchemea/run_chrono.pl',$self->user->name,$session);
    $command .= ' '.$time if defined($time);
    $command .= ' '.$pol if defined($pol);
    $command .= ' &' if (defined($time) && $time == 0); ### Fork if time is 0 (runs forever) ###
    $self->writedebug($command);
    system($command);
    $self->wait(1) if (defined($time) && $time == 0); ### Force wait for 1 minute to allow chronorunforever to start properly before continuing ###
    if ($self->is_locked)
    {
	$self->{'_success'}++;
	return 1;
    }
    return 0;
}

sub potsweep($)
{
    my $self = shift;
    if ($self->{'_test'})
    {
	my $session = shift;
	my @list = $self->user->get_potsweep_points($session);
	my %h = ();
	$self->{'_totaltime'} += $self->pot->sweep_time(\%h,\@list); 
	return;
    }
    $self->{'_count'}++;
    return unless ($self->is_locked);
    $self->writedebug("Starting potsweep");
    my $session = shift;
    $session = $self->user->session unless (defined($session));
    my $string = "Starting chrono current/potential sweep";
    $self->user->proglog($string);
    my $command = join(' ','env DISPLAY=dumb','/usr/local/bin/elchemea/run_potsweep.pl',$self->user->name,$session);
    $self->writedebug($command);
    system($command);
    if ($self->is_locked)
    {
	$self->{'_success'}++;
	return 1;
    }
    return 0;
}

sub multi_imp($$$)
{
    my $self = shift;
    if ($self->{'_test'})
    {
	my $session = shift;
	my $n = shift;
	my $time = shift;
	my @frqlst = $self->user->get_impedance_frequencies($session,$self->pot,$self->fra);
	my $itime = $self->fra->get_impedance_time(\@frqlst) + $self->fra->setuptime + $self->pot->setuptime;
	$self->{'_totaltime'} += $n * $itime + ($n-1) * $time * 60;
	return;
    }
    $self->{'_count'}++;
    return unless ($self->is_locked);
    $self->writedebug("Starting multiimp");
    my $session = shift;
    $session = $self->user->session unless (defined($session));
    my $number = shift;
    my $time = shift;
    my $n = $number-1;
    my $string = "Multiple impedance started";
    $self->user->proglog($string);
    foreach (1..$n)
    {
	$self->impedance($session);
	$self->wait($time);
    }
    $self->impedance($session);
    if ($self->is_locked)
    {
	$self->{'_success'}++;
	return 1;
    }
    return 0;
}

sub impedance($)
{
    my $self = shift;
    if ($self->{'_test'})
    {
	my $session = shift;
	my @frqlst = $self->user->get_impedance_frequencies($session,$self->pot,$self->fra);
	my $itime = $self->fra->get_impedance_time(\@frqlst) + $self->fra->setuptime + $self->pot->setuptime;
	$self->{'_totaltime'} += $itime;
	return;
    }
    $self->{'_count'}++;
    return unless ($self->is_locked);
    $self->writedebug("Starting impedance");
    my $session = shift;
    $session = $self->user->session unless (defined($session));
    my $str = "Starting impedance";
    $self->user->proglog($str);
    my $command = join(' ','env DISPLAY=dumb','/usr/local/bin/elchemea/run_imp.pl',$self->user->name,$session);
    $self->writedebug($command);
    system($command);
    if ($self->is_locked)
    {
	$self->{'_success'}++;
	return 1;
    }
    return 0;
}

sub proglog()
{
    my $self = shift;
    return if ($self->{'_test'});
    $self->{'_count'}++;
    return unless ($self->is_locked);
    my $str = shift;
    chomp $str;
    $self->writedebug($str);
    $self->user->proglog("\# ".$str);
    $self->{'_success'}++;
    return 1;
}

sub change_channel($)
{
    my $self = shift;
    return if ($self->{'_test'});
    $self->{'_count'}++;
    return unless ($self->is_locked);
    my $channel = shift;
    $self->writedebug("Changing multiplexer channel to $channel");
    my $c = undef;
    if ($channel =~ /(\d+)/)
    {
	$c = $1;
    }
    unless (defined($c))
    {
	my $str = "Undefined channel number '$channel' specified for changing multiplexer channel!";
	$self->user->proglog($str);
	EFAlog($str);
	return 0;
    }
    my $string = "Changing multiplexer to channel $c";
    $self->user->proglog($string);
    my $cmd = '/usr/local/bin/elchemea/set_relay.pl '.$c;
    eval {
	system($cmd);
	$self->{'_multiplex'}->writeline($c);
    };
    if ($@)
    {
	my $str = "Error switching to channel $channel: ERROR: $@";
	$self->proglog($str);
	EFAlog($str);
	return 0;
    }
    $self->{'_success'}++;
    return 1;
}

sub get_multiplex_channel()
{
    my $self = shift;
    return if ($self->{'_test'});
    my @lines = $self->{'_multiplex'}->readlines;
    if (@lines && defined($lines[0]) && $lines[0] =~ /(\d+)/)
    {
	return $1;
    }
    return '';
}


sub wait($)
{
    my $self = shift;
    if ($self->{'_test'})
    {
	my $time = shift;
	$self->{'_totaltime'} += $time * 60;
	return;
    }
    $self->{'_count'}++;
    return unless ($self->is_locked);
    my $time = shift;
    $self->writedebug("Waiting $time minutes");
    my $t = time + $time * 60;
    my $string = "Waiting $time minutes";
    $self->user->proglog($string);
    Elchemea::set_proglog($string,$t);
    while (time < $t)
    {
	sleep(1);
	return 0 unless ($self->is_locked);
    }
    $self->{'_success'}++;
    return 1;
}

sub DESTROY
{
    my $self = shift;
    if ($self->{'_test'})
    {
	print $self->{'_totaltime'},"\n";
	return;
    }
    return unless ($self->{'_start_ok'});
    my $string = 'Program finished, completed '.$self->{'_success'}.' out of '.$self->{'_count'}.' commands';
    $self->user->proglog($string);
    $self->unlock; 
}


sub temp_and_ramp($$)
{
    my $self = shift;
    return if ($self->{'_test'});
    my $temp = shift;
    my $ramp = shift;
    $self->{'_count'}++;
    return unless ($self->is_locked);
    my $tty = $self->user->get_config_value('temperature','tty');
    $tty = 'ttyS0' unless ($tty);
    my $channel = $self->user->get_config_value('temperature','address');
    my $comm = $self->user->get_config_value('temperature','controler_mode');
    my $cont = $self->user->get_config_value('temperature','controler_type');
    my $mode = $self->user->get_config_value('temperature','type');
    my $controler = new_controler($cont,$tty,$channel,$comm,$mode);
    unless ($controler)
    {	
	$self->user->proglog("Error: Could not initialise controler '$cont' at tty $tty and address $channel");
	return 0;
    }
    my $res = $controler->set_temp($temp);
    my $string = "Temperature set to $temp";
    $string = "Error: Could not set temperature set to $temp" unless (defined($res));
    $self->user->proglog($string);
    return 0 unless (defined($res));
    $res = $controler->set_ramp($ramp);
    $string = "Temperature set to $ramp C/hour";
    $string = "Error: Could not set temperature set to $ramp C/hour" unless (defined($res));
    $self->user->proglog($string);
    return 0 unless (defined($res));
    $self->{'_success'}++;
    return 1; 
}
